/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Clack.
 * 
 * Clack is free software, also known as "open source."  You can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * (GPL), version 2, as published by the Free Software Foundation (FSF).
 * 
 * Clack is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with Clack; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 * 
 * Clack is modelled on and includes some code fragments from MIT's Click
 * system (http://www.pdos.lcs.mit.edu/click/) which is copyright (c)
 * 1999-2000 Massachusetts Institute of Technology and also GPL'ed.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <stdio.h> // debugging
#include <oskit/io/netio.h>
#include <oskit/io/bufio.h>
#include <clack/c/packet.h>
#ifdef MEASURE
#include <clack/c/stats.h>
#endif

extern oskit_netio_t *nio;

clack_error_t push(clack_packet_t *p);

void freePacket(void* cookie, void* buf) 
{
        clack_packet_t *p = (clack_packet_t*)cookie;
        clack_packet_release(p);
}

clack_error_t
push(clack_packet_t *p)
{
        int rc;
        clack_size_t sz = clack_packet_length(p);
	oskit_bufio_t *b;
#ifdef  MEASURE
	unsigned long long now = get_tsc();
#ifdef  PERFMON
	unsigned long long counter0 = read_perfcounter(0);
	unsigned long long counter1 = read_perfcounter(1);
#endif
#endif
	if (p->bufio) {
		rc = oskit_netio_push(nio, p->bufio, sz);
	}
	else {
		b = oskit_create_extern_bufio(clack_packet_data(p),
					      sz, &freePacket, (void *) p);
		rc = oskit_netio_push(nio, b, sz);
		oskit_bufio_release(b);
	}

#ifdef  MEASURE
	/*
	 * Record time through the router.
	 */
	if (p->measure) {
		if (packet_count >= 100 && packet_flag == 0) {
			packet_flag  = 1;
			packet_count = 0;
			packet_total = 0;
#ifdef  PERFMON
			perfmon_counter0 = 0;
			perfmon_counter1 = 0;
#endif
		}
		packet_count++;
		packet_total += (now - p->created);
#ifdef  PERFMON
		perfmon_counter0 += (counter0 - p->counter0);
		perfmon_counter1 += (counter1 - p->counter1);
#endif
	}
#endif
	if (p->bufio)
		clack_packet_release(p);
        return rc;
}

