/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Clack.
 * 
 * Clack is free software, also known as "open source."  You can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * (GPL), version 2, as published by the Free Software Foundation (FSF).
 * 
 * Clack is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with Clack; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 * 
 * Clack is modelled on and includes some code fragments from MIT's Click
 * system (http://www.pdos.lcs.mit.edu/click/) which is copyright (c)
 * 1999-2000 Massachusetts Institute of Technology and also GPL'ed.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <clack/c/packet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <pcap.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/bpf.h>
#include <clack/c/ethernet.h>
#ifdef MEASURE
#include <clack/c/stats.h>
#endif

/*
 * Imported
 */
extern char		*interface;
extern clack_error_t	push(clack_packet_t *p);

/*
 * Exported
 */
int			fd = { 0 };
void			callback(void);

/*
 * Private.
 */
static pcap_t		*pcap;
static unsigned char	ethaddr[ETHER_ADDR_SIZE];

clack_error_t
init(void)
{
	char		ebuf[PCAP_ERRBUF_SIZE];
	int		one;
	struct	ifreq   ifr;

	printf("frombpf(init): interface=%s\n", interface);

	if (!(pcap = pcap_open_live(interface,
				    2048,	/* XXX snaplen */
				    0,		/* Not in promiscuous mode */
				    1, ebuf))) {
		printf("pcap_open_live failed\n");
		return EINVAL;
	}
	fd  = pcap_fileno(pcap);
	one = 1;
	(void) ioctl(fd, BIOCIMMEDIATE, (caddr_t)&one);

	if (ioctl(fd, SIOCGIFADDR, (caddr_t)&ifr) < 0) {
		printf("FromBPF: Init failed. Could not ioctl(SIOCGIFADDR)\n");
		return errno;
	}
	memcpy(ethaddr, ifr.ifr_addr.sa_data, sizeof(ethaddr));
	
	printf("frombpf(init): fd=%d\n", fd);

	return 0;
}

static void
frompbf_packet_receive(unsigned char *foo,
		       const struct pcap_pkthdr* pkthdr,
		       const unsigned char *data)
{
        clack_packet_t		*p;
	struct ether_header	*e = (struct ether_header *) data;

	/*
	 * If we sent the packet, drop it. Not sure why the BPF
	 * gives us our own packets anyway!
	 */
	if (memcmp(ethaddr, e->ether_shost, sizeof(ethaddr)) == 0)
		return;
	
	p = clack_packet_alloc_padded(pkthdr->caplen, 64 /* XXX */);
	assert(p);

	memcpy(clack_packet_data(p), data, pkthdr->caplen);
#ifdef  MEASURE
	/*
	 * Start clocking ticking after the packet is created since
	 * its pointless to compare the packet creation stuff between
	 * usermode, oskit, and linux.
	 */
	p->created = get_tsc();
	p->measure = 1;
#endif
	push(p);
}

/*
 * This callback gets invoked when the file descriptor has something.
 * Drop into the pcap library and have it hand us all the packets it has
 */ 
void
callback(void)
{
	pcap_dispatch(pcap, -1, frompbf_packet_receive, 0);
}

clack_error_t
fini(void)
{
	pcap_close(pcap);
        return 0;
}
