/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Clack.
 * 
 * Clack is free software, also known as "open source."  You can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * (GPL), version 2, as published by the Free Software Foundation (FSF).
 * 
 * Clack is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with Clack; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 * 
 * Clack is modelled on and includes some code fragments from MIT's Click
 * system (http://www.pdos.lcs.mit.edu/click/) which is copyright (c)
 * 1999-2000 Massachusetts Institute of Technology and also GPL'ed.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

/*
 * ipoutputcombo, which corresponds to the Click combined element used
 * to replace:
 *
 * --> DropBroadcasts --> CheckPaint(1) -->
 *     IPGWOptions(155.99.213.190) -->
 *     FixIPSrc(155.99.213.190) -->
 *     DecIPTTL --> IPFragmenter(1500) --> 
 *
 * with ICMP error generated where appropriate
 */

#include <clack/c/packet.h>
#include <clack/c/packet_ip.h>
#include <arpa/inet.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

/*
 * Imported
 */
extern clack_error_t	push_toip(clack_packet_t *);
extern clack_error_t	push_tofrag(clack_packet_t *);
extern clack_error_t	push_toicmp(clack_packet_t *);
extern struct in_addr   ipaddr;
extern int		color;
extern int		mtu;

/*
 * Exported
 */
clack_error_t		push(clack_packet_t *p);

clack_error_t
push(clack_packet_t *p)
{
	struct ip	*ip  = (struct ip *) clack_packet_data(p);
	int		hlen = ip->ip_hl << 2;
	int		do_cksum = 0;
		
	/* Dropbroadcasts */
	if (p->anno.mac_broadcast)
		goto drop;

	/* check paint */
	if (p->anno.paint == color) {
		goto drop;
	}

	if (hlen > sizeof(struct ip)) {
		puts("ipoutputcombo(push): There are options!\n");
		goto drop;
	}
	
        if (p->anno.fix_ip_src) {
                ip->ip_src = ipaddr;
		p->anno.fix_ip_src = 0;
		do_cksum = 1;
	}

	if (do_cksum) {
		ip->ip_sum = 0;
		ip->ip_sum = in_cksum((unsigned char *)ip, hlen);
	}
	
	/* Decttl */
        if (ip->ip_ttl <= 1) {
		goto drop;
	}
	else {
		unsigned long sum;
		
		ip->ip_ttl--;

		/*
		 * Stole this from Click.
		 *
		 // 19.Aug.1999 - incrementally update IP checksum as suggested by SOSP
		 // reviewers, according to RFC1141, as updated by RFC1624.
		 // new_sum = ~(~old_sum + ~old_halfword + new_halfword)
		 //         = ~(~old_sum + ~old_halfword + (old_halfword - 0x0100))
		 //         = ~(~old_sum + ~old_halfword + old_halfword + ~0x0100)
		 //         = ~(~old_sum + ~0 + ~0x0100)
		 //         = ~(~old_sum + 0xFEFF)
		*/
		sum = (~ntohs(ip->ip_sum) & 0xFFFF) + 0xFEFF;
		ip->ip_sum = ~htons(sum + (sum >> 16));
	}

	/* Checklength(mtu) */
	if (clack_packet_length(p) > mtu) {
		push_tofrag(p);
		return 0;
	}

	/*
	 * Packet okay. Send it!
	 */
	push_toip(p);
	return 0;

	/*
	 * Generate ICMP errors. Not Done yet!
	 */
 drop:
	push_toicmp(p);
	return 0;
}
